

#include <windows.h>
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include <limits.h>

#include <io.h>
#include <string>
#include <vector>
#include <map>

#include "filter.h"
#include "mf17op.h"

// [min, max] ͈̔͂ɖOa
template <class T>
T saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

template <>
int saturate(int v, int min, int max) {
    v -= min;
    v &= ~(v >> 31);
    v += min;

    v -= max;
    v &=  (v >> 31);
    v += max;
    
    return v;
}


static inline int modulo(int a, int b) {
	int m = a % b;

	if (m >= 0) {
		return m;
	} else {
		return m + b;
	}
}

static inline int pixBilinearSub(double x, double y, int v00, int v01, int v10, int v11) {
	return (int)(((v00 * (1.0-x) + v01 * x) * (1.0-y)) + ((v10 * (1.0-x) + v11 * x) * y));
}

static inline PIXEL_YC pixBilinear(PIXEL_YC *buf, int w, int h, int mw, int mh, double x, double y) {
	int zx = (int)x;
	int zy = (int)y;
	double rx = x - (double)zx;
	double ry = y - (double)zy;

	if (rx < 0.0) {
		zx--;
		rx += 1.0;
	}
	if (ry < 0.0) {
		zy--;
		ry += 1.0;
	}

	int zx0 = saturate(zx, 0, w-1);
	int zy0 = saturate(zy, 0, h-1);
	int zx1 = saturate(zx0+1, 0, w-1);
	int zy1 = saturate(zy0+1, 0, h-1);

	PIXEL_YC pix00 = buf[mw * zy0 + zx0];
	PIXEL_YC pix01 = buf[mw * zy0 + zx1];
	PIXEL_YC pix10 = buf[mw * zy1 + zx0];
	PIXEL_YC pix11 = buf[mw * zy1 + zx1];

	PIXEL_YC pix;
	pix.y  = pixBilinearSub(rx,ry, pix00.y,  pix01.y,  pix10.y,  pix11.y);
	pix.cb = pixBilinearSub(rx,ry, pix00.cb, pix01.cb, pix10.cb, pix11.cb);
	pix.cr = pixBilinearSub(rx,ry, pix00.cr, pix01.cr, pix10.cr, pix11.cr);

	return pix;
}


static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;
static std::vector<int> g_bef_check;
static int g_video_rate  = 30;
static int g_video_scale = 1;

static int getTrackVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_trackMap.find(str);

	if (itr == g_trackMap.end()) {
		sprintf_s(errmsg, "track %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	int val = fp->track[itr->second];
	val = std::max<int>(val, fp->track_s[itr->second]);
	val = std::min<int>(val, fp->track_e[itr->second]);

	return val;
}

static int getCheckVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_checkMap.find(str);

	if (itr == g_checkMap.end()) {
		sprintf_s(errmsg, "check %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->check[itr->second];
}

void initFilter(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}
	g_bef_check.resize(fdp->check_n, 0);
}

void updateFilter(FILTER *fp) {
	static int b_period = -1;
	static int b_pdiv = -1;
	static int b_bpm = -1;

	int iperiod = g_trackMap[""];
	int ipdiv = g_trackMap["div"];
	int iexf = g_trackMap["LkF"];
	int ibpm = g_trackMap["bpm*10"];

	int period = getTrackVal(fp, "");
	int pdiv = getTrackVal(fp, "div");
	int exf = getTrackVal(fp, "LkF");
	int bpm = getTrackVal(fp, "bpm*10");

	int video_rate = std::max<int>(g_video_rate, 1);
	int video_scale = std::max<int>(g_video_scale, 1);

	if (b_period != period || b_pdiv != pdiv) {
		bpm = (10 * 60 * video_rate * pdiv) / (video_scale * period);
	} else if (b_bpm != bpm) {
		pdiv = 100;
		period = (10 * 60 * video_rate * pdiv) / (video_scale * bpm);
	}

	int uperiod = std::max<int>(period, 4*exf*pdiv+1);
	if (period < uperiod) {
		period = uperiod;
		bpm = (10 * 60 * video_rate * pdiv) / (video_scale * period);
	}

	b_period = fp->track[iperiod] = period;
	b_pdiv   = fp->track[ipdiv]   = pdiv;
	b_bpm    = fp->track[ibpm]    = bpm;

	fp->exfunc->filter_window_update(fp);

	return;
}


void exitFilter() {
}

struct zoom_inf {
	double rx;
	double ry;
	int end;
};

zoom_inf* getZif(zoom_inf *pzifs, int n, int f) {
	f %= pzifs[n-1].end;
	for (int i=0; i<n; i++) {
		if (f < pzifs[i].end) {
			return (pzifs + i);
		}
	}
	return NULL;
}

BOOL execFilter(FILTER *fp,FILTER_PROC_INFO *fpip) {
	FILE_INFO finf;
	fp->exfunc->get_file_info(fpip->editp, &finf);
	g_video_rate = finf.video_rate;
	g_video_scale = finf.video_scale;

	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;
	int f = fpip->frame;

	int ox = getTrackVal(fp, "x");
	int oy = getTrackVal(fp, "y");
	int period = getTrackVal(fp, "");
	int pdiv = getTrackVal(fp, "div");
	int exf = getTrackVal(fp, "LkF");
	int startf = getTrackVal(fp, "JnF");

	if (f < startf) {
		return TRUE;
	}

	period = std::max<int>(period, 4*exf*pdiv+1);

	static const int N_ZIF = 5;
	zoom_inf zifs[N_ZIF] = {
		{1.0/1.000, 1.0/1.000,   0},
		{1.0/1.000, 1.0/1.015,   0},
		{1.0/1.020, 1.0/0.985,   0},
		{1.0/1.000, 1.0/0.980,   0},
		{1.0/1.000, 1.0/0.990,   0},
	};
	zifs[0].end = period - 4*exf*pdiv;
	zifs[1].end = zifs[0].end + pdiv*exf;
	zifs[2].end = zifs[1].end + pdiv*exf;
	zifs[3].end = zifs[2].end + pdiv*exf;
	zifs[4].end = zifs[3].end + pdiv*exf;


	zoom_inf *pzif = getZif(zifs, N_ZIF, (f - startf)*pdiv);
	if (pzif == NULL) return TRUE;
	if (pzif == zifs) return TRUE;

	double rx = pzif->rx;
	double ry = pzif->ry;

	for (int iy = 0; iy < h; iy++) {
		int src = iy * mw;
		double y = ry * (iy-oy) + oy;
		for (int ix = 0; ix < w; ix++) {
			double x = rx * (ix-ox) + ox;
			buffer1[src] = pixBilinear(buffer0, w,h,mw,mh, x,y);
			src++;
		}
	}
	
	fpip->ycp_edit = buffer1;
	fpip->ycp_temp = buffer0;

	return TRUE;
}
